import unittest
from inspect import getmodule, stack


setup = 0
run = 1
assertion = 2
teardown = 3


def scenario(scenario_name: str):
    """
    Creates a ScenarioBuilder to construct a test case. ScenarioBuilder objects are technically context managers and
    should therefore be used in with constructs:

    ``with scenario("name") as test_scenario:``

    A test scenario can have up to four different parts:

    * ``setup`` *(optional)* - one or more steps to set up the actual test action
    * ``run`` - the action to be taken for the test case execution
    * ``assertion`` *(optional)* - a single assertion for the test case
    * ``teardown`` *(optional)* - one or more steps to clean up after the test case

    All parts take

    :param scenario_name: description of the test case, should be unique since it is converted to a function name
    :return: a ScenarioBuilder object, which can be used to construct test cases
    """
    class ScenarioBuilder:
        class ActionAdder:
            def __init__(self, scenario_builder, action):
                self.scenario_builder = scenario_builder
                self.action = action

            def __lshift__(self, other):
                if self.action == setup:
                    self.scenario_builder.add_setup_action(other)
                if self.action == run:
                    self.scenario_builder.set_run_action(other)
                if self.action == assertion:
                    self.scenario_builder.add_assertion(other)
                if self.action == teardown:
                    self.scenario_builder.add_teardown_action(other)

        def __init__(self, scenario_name_for_builder: str):
            self.scenario_name = scenario_name_for_builder.lower().replace(" ", "_")

            if not self.scenario_name.startswith("test_"):
                self.scenario_name = f"test_{self.scenario_name}"

            self.setup = []
            self.run = None
            self.assertion = None
            self.teardown = []

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            def scenario_execution(_=None):
                list(map(lambda action: action(), self.setup))

                if self.run:
                    self.run()
                else:
                    raise Exception("No run action defined!")

                if self.assertion:
                    self.assertion(_)

                list(map(lambda action: action(), self.teardown))

            class_members = {
                self.scenario_name: scenario_execution
            }

            for key, value in zip(class_members, class_members.values()):
                value.__name__ = key

            module = getmodule(stack()[1][0])

            if "TestSuite" in module.__dict__:
                existing_members = module.__dict__["TestSuite"].__dict__
                for key, value in zip(existing_members, existing_members.values()):
                    class_members[key] = value

            test_class = type("TestSuite", (unittest.TestCase,), class_members)
            module.__dict__["TestSuite"] = test_class

        def __matmul__(self, other):
            return self.ActionAdder(self, other)

        def add_setup_action(self, action):
            self.setup.append(action)

        def set_run_action(self, action):
            if self.run:
                raise Exception("Multiple run actions defined! Only one can be defined per scenario")
            self.run = action

        def add_assertion(self, action):
            if self.assertion:
                raise Exception("Multiple assertions actions defined! Only one can be defined per scenario")
            self.assertion = action

        def add_teardown_action(self, action):
            self.teardown.append(action)

    return ScenarioBuilder(scenario_name)



"""
unittest-specs

A simple wrapper around the unittest package with different styles for specs
"""

__version__ = "1.2.0"
__author__ = "Phillip Goellner"


from unittest_specs.simple_test_spec import SimpleFlatSpec
from unittest_specs.fun_test_spec import describe, it, expect
from unittest_specs.with_test_spec import scenario, setup, run, assertion, teardown



import unittest

from inspect import getmodule, currentframe

from unittest_specs import expect
from unittest_specs.with_test_spec import scenario, setup, run, teardown, assertion


class TestWithScenarioDSL(unittest.TestCase):
    def setUp(self) -> None:
        def increment():
            self.test_value += 1

        self.test_value = 0
        self.increment = increment

    def tearDown(self) -> None:
        module = getmodule(currentframe())
        if "TestSuite" in module.__dict__:
            if "test_scenario" in module.__dict__["TestSuite"].__dict__:
                module.__dict__["TestSuite"] = type("TestSuite", (unittest.TestCase,), {})

    def test_with_statement_registers_class_in_module(self):
        with scenario("test scenario"):
            pass

        module = getmodule(currentframe())
        self.assertEqual(module.__dict__["TestSuite"].__name__, "TestSuite")

    def test_with_statement_registers_scenario_function_in_class(self):
        with scenario("test scenario"):
            pass

        module = getmodule(currentframe())
        self.assertEqual(module.__dict__["TestSuite"].__dict__["test_scenario"].__name__, "test_scenario")

    def test_with_statement_registers_two_scenario_function_in_class(self):
        with scenario("first test scenario"):
            pass

        with scenario("second test scenario"):
            pass

        module = getmodule(currentframe())
        self.assertEqual(module.__dict__["TestSuite"].__dict__["test_first_test_scenario"].__name__, "test_first_test_scenario")
        self.assertEqual(module.__dict__["TestSuite"].__dict__["test_second_test_scenario"].__name__, "test_second_test_scenario")

    def test_execution_throws_exception_when_no_run_action_is_defined(self):
        with scenario("test scenario"):
            pass

        module = getmodule(currentframe())
        test_run = module.__dict__["TestSuite"].__dict__["test_scenario"]

        self.assertRaises(Exception, test_run)

    def test_setup_adds_setup_action_to_test_execution(self):
        with scenario("test scenario") as test_scenario:
            test_scenario @ setup << self.increment
            test_scenario @ run << (lambda: print(""))

        module = getmodule(currentframe())
        module.__dict__["TestSuite"].__dict__["test_scenario"]()
        self.assertEqual(1, self.test_value)

    def test_setup_adds_two_setup_actions_to_test_execution(self):
        with scenario("test scenario") as test_scenario:
            test_scenario @ setup << self.increment
            test_scenario @ setup << self.increment
            test_scenario @ run << (lambda: print(""))

        module = getmodule(currentframe())
        module.__dict__["TestSuite"].__dict__["test_scenario"]()
        self.assertEqual(2, self.test_value)

    def test_run_sets_test_execution_action(self):
        with scenario("test scenario") as test_scenario:
            test_scenario @ run << self.increment

        module = getmodule(currentframe())
        module.__dict__["TestSuite"].__dict__["test_scenario"]()
        self.assertEqual(1, self.test_value)

    def test_setting_a_second_run_action_throws_an_exception(self):
        def create_scenario():
            with scenario("test scenario") as test_scenario:
                test_scenario @ run << self.increment
                test_scenario @ run << self.increment

        self.assertRaises(Exception, create_scenario)

    def test_teardown_adds_teardown_action_to_test_execution(self):
        with scenario("test scenario") as test_scenario:
            test_scenario @ run << (lambda: print(""))
            test_scenario @ teardown << self.increment

        module = getmodule(currentframe())
        module.__dict__["TestSuite"].__dict__["test_scenario"]()
        self.assertEqual(1, self.test_value)

    def test_teardown_adds_two_teardown_actions_to_test_execution(self):
        with scenario("test scenario") as test_scenario:
            test_scenario @ run << (lambda: print(""))
            test_scenario @ teardown << self.increment
            test_scenario @ teardown << self.increment

        module = getmodule(currentframe())
        module.__dict__["TestSuite"].__dict__["test_scenario"]()
        self.assertEqual(2, self.test_value)


with scenario("free standing scenario") as standalone_test_scenario:

    class Adder:
        def __init__(self):
            self.__result = 0

        def perform_addition(self, a, b):
            self.__result = a + b

        def reset_result(self):
            self.__result = 0

        @property
        def result(self):
            return self.__result

    adder = Adder()

    standalone_test_scenario @ setup << (lambda: adder.reset_result())
    standalone_test_scenario @ run << (lambda: adder.perform_addition(1, 4))
    standalone_test_scenario @ assertion << expect(lambda: adder.result).to_be(5)
    standalone_test_scenario @ teardown << (lambda: adder.reset_result())
